/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#pragma once

#define LOG_SENDER_FRAME_DROPS false

#include "ExpFrameSender.h"
#include "MixCastSdkActor.h"
#include "Utility/MathUtil.hpp"

TMap<FString, SharedQueueProducer<ExpFrame>*> FExpFrameSender::cachedFrameQueues;

void FExpFrameSender::DisposeAllCachedQueues()
{
	for (auto& Elem : cachedFrameQueues)
		delete Elem.Value;
	cachedFrameQueues.Reset();
}

void FExpFrameSender::PopulateLights(ExpLight lights[])
{
	FTransform trackingRootInv = AMixCastSdkActor::GetWorldTransform().Inverse();
	float worldToMeters = cam->GetWorld()->GetWorldSettings()->WorldToMeters;

	int listIndex = 0;
	int arrIndex = 0;
	while (arrIndex < ExpFrame::MaxLights)
	{
		if (listIndex >= UMixCastLightComponent::Active.Num())
		{
			lights[arrIndex++].type = ExpLightType::None;
			continue;
		}

		ULightComponent* light = UMixCastLightComponent::Active[listIndex++]->GetSource();
		if (light->Intensity <= 0)
			continue;

		lights[arrIndex].color = FVector(light->LightColor.R, light->LightColor.G, light->LightColor.B) * light->Intensity;

		lights[arrIndex].pos = MathUtil::UnrealPositionToMixCast(trackingRootInv.TransformPosition(light->GetComponentLocation()), worldToMeters);
		lights[arrIndex].forward = MathUtil::UnrealVectorToMixCast(trackingRootInv.TransformVector(light->GetForwardVector()));
		lights[arrIndex].forward.Normalize();

		lights[arrIndex].range = 0;
		lights[arrIndex].angle = 0;

		UClass* lightClass = light->GetClass();
		if (lightClass == UPointLightComponent::StaticClass()) {
			lights[arrIndex].type = ExpLightType::Point;
			lights[arrIndex].range = ((UPointLightComponent*)light)->AttenuationRadius / worldToMeters;
		}
		else if (lightClass == UDirectionalLightComponent::StaticClass()) {
			lights[arrIndex].type = ExpLightType::Directional;
		}
		else if (lightClass == USpotLightComponent::StaticClass()) {
			lights[arrIndex].type = ExpLightType::Spot;
			lights[arrIndex].range = ((USpotLightComponent*)light)->AttenuationRadius / worldToMeters;
			lights[arrIndex].angle = FMath::DegreesToRadians(((USpotLightComponent*)light)->OuterConeAngle);
		}
		else
			continue;

		arrIndex++;
	}
}

void FExpFrameSender::CreateQueue()
{
	FString queueId = SharedTexIds::Cameras::RenderedExpFrames::Get(cam->Identifier);
	SharedQueueProducer<ExpFrame>** foundQueue = cachedFrameQueues.Find(queueId);
	if (foundQueue == nullptr)
	{
		std::wstring srcId = StringUtils::ToStdWString(queueId);
		renderedExpFrames = new SharedQueueProducer<ExpFrame>(srcId.c_str(), QueueCapacity);
		cachedFrameQueues.Add(queueId, renderedExpFrames);
	}
	else
		renderedExpFrames = *foundQueue;
}
void FExpFrameSender::CreateTextureList()
{
	sharedExpLayersTexs = new OwnedUnrealSharedTexture * [QueueCapacity];
	for (int i = 0; i < QueueCapacity; i++)
		sharedExpLayersTexs[i] = NULL;

	sharedExpOpaqueLayerTexs = new OwnedUnrealSharedTexture * [QueueCapacity];
	for (int i = 0; i < QueueCapacity; i++)
		sharedExpOpaqueLayerTexs[i] = NULL;
}
void FExpFrameSender::DestroyTextureList()
{
	if (sharedExpLayersTexs != NULL)
	{
		for (int i = 0; i < QueueCapacity; i++) {
			if (sharedExpLayersTexs[i] != NULL) {
				sharedExpLayersTexs[i]->BeginDestroy();
				sharedExpLayersTexs[i] = NULL;
			}
		}
		delete[] sharedExpLayersTexs;
		sharedExpLayersTexs = NULL;
	}

	if (sharedExpOpaqueLayerTexs != NULL)
	{
		for (int i = 0; i < QueueCapacity; i++) {
			if (sharedExpOpaqueLayerTexs[i] != NULL) {
				sharedExpOpaqueLayerTexs[i]->BeginDestroy();
				sharedExpOpaqueLayerTexs[i] = NULL;
			}
		}
		delete[] sharedExpOpaqueLayerTexs;
		sharedExpOpaqueLayerTexs = NULL;
	}
}

FExpFrameSender::FExpFrameSender(AExpCamera* cam)
{
	this->cam = cam;

	CreateQueue();
	CreateTextureList();

	frameDelegate = AExpCamera::OnFrameEnded().AddRaw(this, &FExpFrameSender::HandleFrameUpdated);
}
FExpFrameSender::~FExpFrameSender()
{
	AExpCamera::OnFrameEnded().Remove(frameDelegate);

	DestroyTextureList();
}

void FExpFrameSender::HandleFrameUpdated(AExpCamera* curCam)
{
	if (curCam != cam)
		return;

	renderedExpFrames->EnqueueFenceUpdates();

	ExpFrame renderedFrame = cam->GetLatestFrameInfo();

	UMixCastSliceCanvas* layersCanvas = cam->GetSliceCanvas();
	if (!layersCanvas->IsValidLowLevel() || !layersCanvas->Resource->TextureRHI || !layersCanvas->Resource->TextureRHI->GetNativeResource())
		return;

#if MIXCAST_SHADER_PASSES
	UMixCastOpaqueCanvas* opaqueCanvas = cam->GetOpaqueCanvas();
	UTextureRenderTarget2D* opaqueLayerTex = NULL;

	if (opaqueCanvas != NULL)
	{
		if (renderedFrame.HasCamFlag(ExpCamFlagBit::SeparateOpaque) &&
			(!opaqueCanvas || !opaqueCanvas->IsValidLowLevel() || !opaqueCanvas->Resource->TextureRHI || !opaqueCanvas->Resource->TextureRHI->GetNativeResource() || !opaqueCanvas->Resource->IsInitialized()))
			return;
	}
	else
	{
		opaqueLayerTex = cam->GetHdrOpaqueLayerTarget();
		if (renderedFrame.HasCamFlag(ExpCamFlagBit::SeparateOpaque) &&
			(!opaqueLayerTex || !opaqueLayerTex->IsValidLowLevel() || !opaqueLayerTex->Resource->TextureRHI || !opaqueLayerTex->Resource->TextureRHI->GetNativeResource()))
			return;
	}
#endif

	if (!renderedExpFrames->WaitUntilNextEmptied(0))
	{
		if (!consumerMissedLastFrame && firstWriteCompleted && LOG_SENDER_FRAME_DROPS)
		{
			MIXCAST_WARNING("Compositing process started missing frames at time {%s}",
				*TimeUtils::GetSMPTE(TimeUtils::GetMidnightRelativeFrameIndex(renderedFrame.syncTime, renderedFrame.camFramerate), renderedFrame.camFramerate));
		}
		consumerMissedLastFrame = true;
		return;
	}

	//Copy layer tex
	OwnedUnrealSharedTexture* sharedTex = sharedExpLayersTexs[renderedExpFrames->GetCurrentSlot()];
	if (sharedTex == NULL)
	{
		sharedTex = OwnedUnrealSharedTexture::Create(layersCanvas);
		sharedExpLayersTexs[renderedExpFrames->GetCurrentSlot()] = sharedTex;
	}
	sharedTex->EnqueueWrite(layersCanvas);
	renderedFrame.layersTex = sharedTex->GetHandle();

#if MIXCAST_SHADER_PASSES
	if (renderedFrame.HasCamFlag(ExpCamFlagBit::SeparateOpaque))
	{
		UTexture* opaqueTex = opaqueCanvas ? opaqueCanvas : opaqueLayerTex;
		sharedTex = sharedExpOpaqueLayerTexs[renderedExpFrames->GetCurrentSlot()];
		if (sharedTex == NULL)
		{
			sharedTex = OwnedUnrealSharedTexture::Create(opaqueTex);
			sharedExpOpaqueLayerTexs[renderedExpFrames->GetCurrentSlot()] = sharedTex;
		}
		sharedTex->EnqueueWrite(opaqueTex);
		renderedFrame.opaqueLayerTex = sharedTex->GetHandle();
	}
	else
		renderedFrame.opaqueLayerTex = NULL;
#endif

	PopulateLights(renderedFrame.lights);

	renderedExpFrames->Write(renderedFrame, true);

	if (!firstWriteCompleted && LOG_SENDER_FRAME_DROPS)
	{
		firstWriteCompleted = true;
		consumerMissedLastFrame = false;

		MIXCAST_LOG("Compositing process started at time {%s}",
			*TimeUtils::GetSMPTE(TimeUtils::GetMidnightRelativeFrameIndex(renderedFrame.syncTime, renderedFrame.camFramerate), renderedFrame.camFramerate));
	}

	if (consumerMissedLastFrame && LOG_SENDER_FRAME_DROPS)
	{
		MIXCAST_LOG("Compositing process back in business at time {%s}",
			*TimeUtils::GetSMPTE(TimeUtils::GetMidnightRelativeFrameIndex(renderedFrame.syncTime, renderedFrame.camFramerate), renderedFrame.camFramerate));
	}

	consumerMissedLastFrame = false;
}