/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "PluginProjectSettings.h"

#include "MixCastLog.h"
#include "Capture/ExpCamera.h"
#include "Display/Projections/ExpVideoInput.h"
#include "Display/Viewfinders/ExpViewfinder.h"

#if UE_EDITOR
#include "ISettingsModule.h"
#include "ISettingsSection.h"
#include "ISettingsContainer.h"

#include "Modules/ModuleManager.h"
#include "Misc/ConfigCacheIni.h"
#include "HAL/ConsoleManager.h"
#include "Engine/RendererSettings.h"
#endif

#define LOCTEXT_NAMESPACE "MixCast"

UPluginProjectSettings::UPluginProjectSettings(const FObjectInitializer& ObjectInitializer) : Super(ObjectInitializer)
{
	if (CameraClass.IsNull())
		CameraClass = FStringClassReference(AExpCamera::StaticClass());
	if (VideoInputClass.IsNull())
		VideoInputClass = FStringClassReference(AExpVideoInput::StaticClass());
	if (ViewfinderClass.IsNull())
		ViewfinderClass = FStringClassReference(AExpViewfinder::StaticClass());
}

void UPluginProjectSettings::ValidateCurrentSettings()
{
	const int RequiredResizeMethod = 2;
	static IConsoleVariable* SceneRenderTargetResizeMethodCVar = IConsoleManager::Get().FindConsoleVariable(TEXT("r.SceneRenderTargetResizeMethod"));
	if (SceneRenderTargetResizeMethodCVar && SceneRenderTargetResizeMethodCVar->GetInt() != RequiredResizeMethod)
		SceneRenderTargetResizeMethodCVar->Set(RequiredResizeMethod);

#if MIXCAST_SHADER_PASSES
	if (ClipMode == EMixCastClipMode::Planar)
	{
		ClipMode = EMixCastClipMode::PerPixel;

		Modify();
		SaveConfig();
	}
#else
	if (ClipMode == EMixCastClipMode::PerPixel && TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
	{
		MIXCAST_WARNING("Per-Pixel Clip Mode doesn't support Translucency from Alpha! Changing to Planar Clip Mode");

		ClipMode = EMixCastClipMode::Planar;

		Modify();
		SaveConfig();

#if UE_EDITOR
		EnforceGlobalClipPlaneSettingIfNeeded();
#endif
	}
	if (DoesntProduceGoodDepth == false)
	{
		DoesntProduceGoodDepth = true;

		Modify();
		SaveConfig();
	}
#endif
}

#if UE_EDITOR
void UPluginProjectSettings::Register()
{
	if (ISettingsModule* SettingsModule = FModuleManager::GetModulePtr<ISettingsModule>("Settings"))
	{
		ISettingsContainerPtr SettingsContainer = SettingsModule->GetContainer("Project");
		ISettingsSectionPtr SettingsSection = SettingsModule->RegisterSettings("Project", "Plugins", "MixCast",
			LOCTEXT("RuntimeGeneralSettingsName", "MixCast"),
			LOCTEXT("RuntimeGeneralSettingsDescription", "Project configuration for the MixCast SDK"),
			GetMutableDefault<UPluginProjectSettings>()
		);
	}
}
void UPluginProjectSettings::Unregister()
{
	if (ISettingsModule* SettingsModule = FModuleManager::GetModulePtr<ISettingsModule>("Settings"))
	{
		SettingsModule->UnregisterSettings("Project", "Plugins", "MixCast");
	}
}

bool UPluginProjectSettings::CanEditChange(const UPROPCLASS* InProperty) const
{
	if (!Super::CanEditChange(InProperty))
		return false;

	FName PropertyName = InProperty->GetFName();
	if (PropertyName == GET_MEMBER_NAME_CHECKED(UPluginProjectSettings, ClipMode))
	{
#if MIXCAST_SHADER_PASSES
		return false;
#else
		if (TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
			return false;
#endif
	}
#if !MIXCAST_SHADER_PASSES
	else if (PropertyName == GET_MEMBER_NAME_CHECKED(UPluginProjectSettings, DoesntProduceGoodDepth))
	{
		return false;
	}
#endif
	return true;
}

#if !MIXCAST_SHADER_PASSES
void UPluginProjectSettings::PostEditChangeProperty(struct FPropertyChangedEvent& PropertyChangedEvent)
{
	Super::PostEditChangeProperty(PropertyChangedEvent);

	FName PropertyName = (PropertyChangedEvent.MemberProperty != NULL) ? PropertyChangedEvent.MemberProperty->GetFName() : NAME_None;
	if (PropertyName == GET_MEMBER_NAME_CHECKED(UPluginProjectSettings, TranslucencyMode))
	{
		if (ClipMode == EMixCastClipMode::PerPixel && TranslucencyMode == EMixCastTranslucencyMode::FromAlpha)
		{
			MIXCAST_WARNING("Per-Pixel Clip Mode doesn't support Translucency from Alpha! Changing to Planar Clip Mode");

			ClipMode = EMixCastClipMode::Planar;

			Modify();

			EnforceGlobalClipPlaneSettingIfNeeded();
		}
	}
	else if (PropertyName == GET_MEMBER_NAME_CHECKED(UPluginProjectSettings, ClipMode))
	{
		EnforceGlobalClipPlaneSettingIfNeeded();
	}
#if !MIXCAST_SHADER_PASSES
	else if (PropertyName == GET_MEMBER_NAME_CHECKED(UPluginProjectSettings, DoesntProduceGoodDepth))
	{
		if (DoesntProduceGoodDepth == false)
		{
			DoesntProduceGoodDepth = true;
			Modify();
		}
	}
#endif
}
void UPluginProjectSettings::EnforceGlobalClipPlaneSettingIfNeeded()
	{
		if (ClipMode == EMixCastClipMode::Planar)
		{
			if (GConfig->IsReadyForUse())
			{
			URendererSettings* settings = GetMutableDefault<URendererSettings>();
			if (settings->bGlobalClipPlane != 1) {
				MIXCAST_WARNING("Setting Global Clip Plane setting as required by Planar Clip Mode. You'll need to restart the editor before MixCast will function correctly");

				settings->bGlobalClipPlane = 1;
				GConfig->SetBool(
					TEXT("/Script/Engine.RendererSettings"),
					TEXT("r.AllowGlobalClipPlane"),
					true,
					GEngineIni);
				GConfig->Flush(false, GEngineIni);
			}
		}
	}
}
#endif
#endif

#undef LOCTEXT_NAMESPACE
