/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "ExpVideoInput.h"

#include "MixCast.h"
#include "Capture/ExpCamera.h"
#include "Transmission/Generated/VirtualCamera_types.h"
#include "Data/SharedTexIds.h"

#include "Engine/Engine.h"
#include "UObject/ConstructorHelpers.h"
#include "Engine/Texture2D.h"
#include "GameFramework/WorldSettings.h"

#include <string>

AExpVideoInput::AExpVideoInput(const FObjectInitializer &ObjectInitializer)
{
	static ConstructorHelpers::FObjectFinder<UMaterial> Find_ProjectVideoMat(TEXT("Material'/MixCast/ProjectVideo.ProjectVideo'"));
	if (Find_ProjectVideoMat.Succeeded())
		ProjectionMatTemplate = Find_ProjectVideoMat.Object;

	SpawnCollisionHandlingMethod = ESpawnActorCollisionHandlingMethod::AlwaysSpawn;

	VideoInputTransform = CreateDefaultSubobject<USceneComponent>(FName("Transform"));
	SetRootComponent(VideoInputTransform);

	ProjectionMesh = CreateDefaultSubobject<UProceduralMeshComponent>(FName("Projection Mesh"));
	ProjectionMesh->SetupAttachment(VideoInputTransform);

	PrimaryActorTick.bCanEverTick = true;
	PrimaryActorTick.TickGroup = TG_PostUpdateWork;
	PrimaryActorTick.bStartWithTickEnabled = true;
}

void AExpVideoInput::BeginPlay()
{
	ColorReceiver = new SharedTextureReceiver(SharedTexIds::VideoInputs::LatestColor::Get(Identifier));
	ColorReceiver->OnTextureReferenceChanged().AddUObject(this, &AExpVideoInput::HandleColorTextureChanged);

	DepthReceiver = new SharedTextureReceiver(SharedTexIds::VideoInputs::LatestDepth::Get(Identifier));
	DepthReceiver->OnTextureReferenceChanged().AddUObject(this, &AExpVideoInput::HandleDepthTextureChanged);

	Super::BeginPlay();

	ProjectionMatInstance = UMaterialInstanceDynamic::Create(
		Cast<UMaterialInterface>(ProjectionMatTemplate),
		this,
		TEXT("Project VideoInput")
	);
	ProjectionMesh->SetMaterial(0, ProjectionMatInstance);

	HandleColorTextureChanged();
	HandleDepthTextureChanged();

	FrameStartedDelegate = AExpCamera::OnFrameStarted().AddUObject(this, &AExpVideoInput::HandleFrameStarted);
}
void AExpVideoInput::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	AExpCamera::OnFrameStarted().Remove(FrameStartedDelegate);

	if (ProjectionMesh) {
		if (ProjectionMesh->IsValidLowLevel()) {
			ProjectionMesh->DestroyComponent();
		}
		ProjectionMesh = nullptr;
	}

	if (ColorReceiver)
	{
		ColorReceiver->OnTextureReferenceChanged().AddUObject(this, &AExpVideoInput::HandleColorTextureChanged);
		delete ColorReceiver;
		ColorReceiver = nullptr;
	}
	if (DepthReceiver)
	{
		DepthReceiver->OnTextureReferenceChanged().AddUObject(this, &AExpVideoInput::HandleDepthTextureChanged);
		delete DepthReceiver;
		DepthReceiver = nullptr;
	}

	Super::EndPlay(EndPlayReason);
}

void AExpVideoInput::HandleFrameStarted(AExpCamera* cam)
{
	VideoInput* viInfo = SdkData::GetVideoInputById(Identifier);
	if (viInfo == nullptr)
		return;

	bool displayToCamera = false;
	for (int i = 0; i < viInfo->projectToCameras.size() && !displayToCamera; i++)
		displayToCamera |= viInfo->projectToCameras[i] == cam->Identifier;

	if (!displayToCamera)
		cam->SetActorHidden(this);
}

void AExpVideoInput::HandleColorTextureChanged()
{
	ProjectionMatInstance->SetTextureParameterValue(FName("MainTex"), ColorReceiver->GetTexture());
}
void AExpVideoInput::HandleDepthTextureChanged()
{
	ProjectionMatInstance->SetTextureParameterValue(FName("DepthTex"), DepthReceiver->GetTexture());
}

void AExpVideoInput::Tick(float DeltaSeconds)
{
	VideoInput* videoInputData = SdkData::GetVideoInputById(Identifier);
	if (videoInputData == nullptr)
		return;

	float worldToMeters = GetWorld()->GetWorldSettings()->WorldToMeters;

	FVector camPos = ThriftUtil::ThriftPositionToUnreal(videoInputData->currentPosition, worldToMeters);
	FQuat camRot = ThriftUtil::ThriftQuatToUnreal(videoInputData->currentRotation);
	RootComponent->SetRelativeLocationAndRotation(camPos, camRot);
	RootComponent->UpdateComponentToWorld();
	 
	const UTexture2D* colorTex = ColorReceiver->GetTexture();
	const UTexture2D* depthTex = DepthReceiver->GetTexture();

	bool drawable = colorTex != nullptr && depthTex != nullptr;
	ProjectionMesh->SetVisibility(drawable);
	if (drawable)
	{
		int newCellsX = FMath::Max(2, depthTex->GetSizeX()) - 1;
		int newCellsY = FMath::Max(2, depthTex->GetSizeY()) - 1;

		if (newCellsX != lastMeshWidth || newCellsY != lastMeshHeight)
			RegenerateMesh(newCellsX, newCellsY);

		ColorReceiver->EnqueueUpdate();
		DepthReceiver->EnqueueUpdate();

		float yProj = FMath::Tan(FMath::DegreesToRadians((float)videoInputData->fieldOfView * 0.5f));
		float xProj = yProj * ColorReceiver->GetTexture()->GetSizeX() / ColorReceiver->GetTexture()->GetSizeY();
		ProjectionMatInstance->SetVectorParameterValue(FName("ProjectionExtents"), FLinearColor(xProj, yProj, 0, 0));
	}
}

void AExpVideoInput::RegenerateMesh(int cellsX, int cellsY)
{
	ProjectionMesh->ClearAllMeshSections();
	//ProjectionMesh->CreateMeshSection()
	TArray<FVector> vertices;
	TArray<FVector> normals;
	TArray<FVector2D> uvs;
	TArray<int32> indices;
	TArray<FColor> colors;
	TArray<FProcMeshTangent> tangents;

	for (int y = 0; y <= cellsY; y++)
	{
		float v = (float)y / cellsY;
		for (int x = 0; x <= cellsX; x++)
		{
			float u = (float)x / cellsX;

			vertices.Add(2 * FVector(0, u - 0.5f, v - 0.5f));
			normals.Add(FVector::UpVector);
			tangents.Add(FProcMeshTangent(0, 1, 0));
			colors.Add(FColor::Black);
			uvs.Add(FVector2D(u, v));
		}
	}	
	for (int y = 0; y < cellsY; y++)
	{
		for (int x = 0; x < cellsX; x++)
		{
			int bottomLeft = y * (cellsX + 1) + x;
			int bottomRight = bottomLeft + 1;
			int topLeft = (y + 1) * (cellsX + 1) + x;
			int topRight = topLeft + 1;

			//Alternate how quads are broken into triangles per row
			if (y % 2 == 0) {
				indices.Add(bottomLeft);
				indices.Add(topRight); 
				indices.Add(topLeft);

				indices.Add(bottomLeft);
				indices.Add(bottomRight); 
				indices.Add(topRight);
			}
			else {
				indices.Add(bottomRight);
				indices.Add(topLeft); 
				indices.Add(bottomLeft);

				indices.Add(bottomRight);
				indices.Add(topRight);
				indices.Add(topLeft);
			}
		}
	}

	ProjectionMesh->CreateMeshSection(0, vertices, indices, normals, uvs, colors, tangents, false);

	//Reassign bounds (set after CreateMeshSection) to enormous to ensure Frustum Culling doesn't cull projection unintentionally
	ProjectionMesh->Bounds.BoxExtent = FVector(10000, 10000, 10000);
	ProjectionMesh->Bounds.SphereRadius = 10000;
	ProjectionMesh->BoundsScale = 10000;

	lastMeshWidth = cellsX;
	lastMeshHeight = cellsY;
}

UTexture2D* AExpVideoInput::GetColorTexture()
{
	return ColorReceiver->GetTexture();
}
UTexture2D* AExpVideoInput::GetDepthTexture()
{
	return DepthReceiver->GetTexture();
}