#pragma once

#include "Utility/ProjectUtils.hpp"

#if PLATFORM_WINDOWS
#include "Windows/AllowWindowsPlatformTypes.h"
#endif
#include <Windows.h>
#include <stdio.h>
#include <conio.h>
#include <d3d11.h>
#if PLATFORM_WINDOWS
#include "Windows/HideWindowsPlatformTypes.h"
#endif

#include "RHICommandList.h"

template <class T>
class SharedQueueReader
{
private:
	const int KEY_SIZE = 256;

	int dataSize;

	HANDLE filledEvent = NULL;
	HANDLE emptiedEvent = NULL;

	HANDLE dataFile = NULL;
	void* dataPtr = NULL;

#if MIXCAST_D3D12
	bool useEngineFence = false;
	FGPUFenceRHIRef fence;
#endif

	ID3D11Device* fenceDev;
	ID3D11DeviceContext* fenceContext = NULL;
	ID3D11Query* fenceQuery = NULL;

	void InsertEmptiedFence(FRHICommandListImmediate& RHICmdList)
	{
#if MIXCAST_D3D12
		if (useEngineFence) 
		{
			fence = RHICmdList.CreateGPUFence(FName("Queue Read"));
			RHICmdList.WriteGPUFence(fence);
		}
		else
		{
#endif
			D3D11_QUERY_DESC queryDesc;
			queryDesc.Query = D3D11_QUERY_EVENT;
			queryDesc.MiscFlags = 0;
			fenceDev->CreateQuery(&queryDesc, &fenceQuery);

			fenceDev->GetImmediateContext(&fenceContext);
			fenceContext->End(fenceQuery);
#if MIXCAST_D3D12
		}
#endif
	}
	void UpdateFence()
	{
#if MIXCAST_D3D12
		if (useEngineFence) 
		{
			if (!fence.IsValid())
				return;

			if (fence->Poll())
				MarkEmptied();
		}
		else
		{
#endif
			if (fenceQuery == NULL)
				return;

			if (fenceContext->GetData(fenceQuery, NULL, 0, 0) == S_OK)
				MarkEmptied();
#if MIXCAST_D3D12
		}
#endif
	}

public:
	//Unfenced
	void MarkEmptied()
	{
#if MIXCAST_D3D12
		fence.SafeRelease();
#endif
		SAFE_RELEASE(fenceContext);
		SAFE_RELEASE(fenceQuery);

		SetEvent(emptiedEvent);
	}


	SharedQueueReader(LPCWSTR frameId)
	{
		dataSize = sizeof(T);

#if MIXCAST_D3D12
		useEngineFence = ProjectUtils::GetGraphicsType() != ExpGraphicsType::DX11;
		if (!useEngineFence)
#endif
			fenceDev = (ID3D11Device*)GDynamicRHI->RHIGetNativeDevice();

		LPWSTR key = new TCHAR[KEY_SIZE];

		//Events
		wsprintf(key, TEXT("Global\\MixCast.SharedQ(%s).Filled"), frameId);
		filledEvent = CreateEvent(NULL, true, false, key);
		wsprintf(key, TEXT("Global\\MixCast.SharedQ(%s).Emptied"), frameId);
		emptiedEvent = CreateEvent(NULL, true, false, key);

		//Metadata
		wsprintf(key, TEXT("MixCast-SharedQ(%s)-Data"), frameId);
		dataFile = OpenFileMapping(FILE_MAP_READ, false, key);
		dataPtr = (void*)MapViewOfFile(dataFile,   // handle to map object
			FILE_MAP_READ, // read/write permission
			0,
			0,
			dataSize);

		delete[] key;

		MarkEmptied();
	}
	void BeginDestroy()
	{
#if MIXCAST_D3D12
		fence.SafeRelease();
#endif
		SAFE_RELEASE(fenceContext);
		SAFE_RELEASE(fenceQuery);

		UnmapViewOfFile(dataPtr);
		CloseHandle(dataFile);

		CloseHandle(filledEvent);
		CloseHandle(emptiedEvent);

		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
			{
				delete this;
			});
	}


	bool WaitUntilFilled(int timeout)
	{
		if (timeout == -1) {
			WaitForSingleObject(filledEvent, INFINITE);
			ResetEvent(filledEvent);
			return true;
		}
		else {
			DWORD res = WaitForSingleObject(filledEvent, timeout);
			if (res == WAIT_OBJECT_0)
			{
				ResetEvent(filledEvent);
				return true;
			}
			return false;
		}
	}
	void Read(T& dst)
	{
		CopyMemory(&dst, dataPtr, dataSize);
	}

	void EnqueueFenceInsert()
	{
		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
		{
			InsertEmptiedFence(RHICmdList);
		});
	}
	void EnqueueFenceUpdate()
	{
		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
		{
			UpdateFence();
		});
	}
};

